;;;; miscmacros.scm

;;; Modify locations, T-like:

(cond-expand (syntax-case

(define-syntax (modify-location x)
  (syntax-case x ()
    ((_ (loc ...) proc)
     (with-syntax (((tmp ...) (generate-temporaries #'(loc ...))))
       #'(let ((tmp loc) ...)
           (proc (lambda () (tmp ...))
                 (lambda (x) (set! (tmp ...) x)))) ) )
    ((_ loc proc)
     #'(proc (lambda () loc)
             (lambda (x) (set! loc x)) ) ) ) )

)(else

(define-macro (modify-location loc proc)
  (let ((x (gensym)))
    (if (atom? loc)
        `(,proc (lambda () ,loc)
                (lambda (,x) (set! ,loc ,x)) ) 
        (let ((tmps (map (lambda _ (gensym)) loc)))
          `(let ,(map list tmps loc)
             (,proc (lambda () ,tmps)
                    (lambda (,x) (set! ,tmps ,x))))) ) ) )

) )

;; evaluates body with an explicit exit continuation
;;
(define-macro (let/cc K . BODY)
  `(call-with-current-continuation (lambda (,K) ,@BODY)))

;; loop while expression false
;;
(define-macro (until TEST . BODY)
  (let ([LOOP-TAG (gensym)])
    `(let ,LOOP-TAG ()
      (unless ,TEST
        ,@BODY
        (,LOOP-TAG)))))

;; repeat body n times
;;
(define-macro (repeat N . BODY)
  (let ([CTR-VAR (gensym)]
        [LOOP-TAG (gensym)])
    `(let ,LOOP-TAG ([,CTR-VAR ,N])
      (when (< 0 ,CTR-VAR)
        ,@BODY
        (,LOOP-TAG (sub1 ,CTR-VAR))))))

;; repeat body n times, w/ countdown n bound to 'it'
;;
(define-macro (repeat* N . BODY)
  (let ([LOOP-TAG (gensym)])
    `(let ,LOOP-TAG ([it ,N])
      (when (< 0 it)
        ,@BODY
        (,LOOP-TAG (sub1 it))))))

;; repeat body n times, w/ countup n bound to "var"
;;
;; NVF: (<variable> <expression> [<finally>])
;;
(define-macro (dotimes NVF . BODY)
  (if (and (list? NVF) (<= 2 (length NVF) 3))
		(let ([V (car NVF)]
					[N (cadr NVF)]
					[F (if (= 3 (length NVF)) (caddr NVF) '(begin))]
					[N-VAR (gensym)]
					[LOOP-TAG (gensym)])
			`(let ,LOOP-TAG ([,V 0] [,N-VAR ,N])
				(cond [(< ,V ,N-VAR)
					     ,@BODY
					     (,LOOP-TAG (add1 ,V) ,N-VAR)]
					    [else
					      ,F] ) ) )
		(syntax-error 'dotimes "not a binding form" NVF) ) )

(define-macro (while x . xs)
  (let ([var (gensym)])
    `(let ,var ()
        (when ,x (let () ,@xs (,var))) ) ) )

(define-macro (while* x . xs)
  (let ([LBL (gensym)])
    `(let ,LBL ()
        (if* ,x (let () ,@xs (,LBL))) ) ) )

(define-macro (if* x y . z)
  (let ([var (gensym)])
    `(let ([,var ,x])
       (if ,var
           (let ([it ,var])
             ,y)
           ,@z) ) ) )

(define-macro (push! x y)
  `(modify-location 
    ,y 
    (lambda (get set) (set (cons ,x (get))) ) ) )

(define-macro (pop! x)
  (let ([var (gensym)]
        (var2 (gensym)) )
    `(modify-location
      ,x
      (lambda (get set)
        (let* ([,var (get)]
               (,var2 (car ,var)) )
          (set (cdr ,var))
          ,var2) ) ) ) )

(define-macro (ignore-errors . xs)
  `(handle-exceptions _ #f ,@xs) )

(define-macro (begin0 x1 . xs)
  (let ([var (gensym)])
    `(##sys#call-with-values
      (lambda () ,x1)
      (lambda ,var 
        (begin ,@xs (apply ##sys#values ,var) ) ) ) ) )

(define-macro (define-optionals vars args)
  `(begin
     ,@(map (lambda (b) `(define ,(car b) #f)) vars)
     ,(let ([aliases (map (lambda (b) (gensym (car b))) vars)])
        `(let-optionals ,args ,(map (lambda (b a) (cons a (cdr b))) vars aliases)
           ,@(map (lambda (b a) `(set! ,(car b) ,a)) vars aliases) ) ) ) )

(define-macro (define-parameter name . more)
  (let-optionals* more ([init '(void)] more)
    `(define ,name
       (make-parameter ,init ,@more) ) ) )

(define-macro (inc! v)
  `(modify-location
    ,v
    (lambda (get set) (set (add1 (get))))))

(define-macro (dec! v)
  `(modify-location
    ,v
    (lambda (get set) (set (sub1 (get))))))

(define-macro (exchange! x y)
  (let ((tmp (gensym))
	(g1 (gensym))
	(g2 (gensym))
	(s1 (gensym))
	(s2 (gensym)) )
    `(modify-location
      ,x
      (lambda (,g1 ,s1)
	(modify-location
	 ,y
	 (lambda (,g2 ,s2)
	   (let ((,tmp (,g1)))
	     (,s1 (,g2))
	     (,s2 ,tmp) ) ) ) ) ) ) )

(define-macro (modify! x proc)
  (let ((get (gensym))
	(set (gensym)) )
    `(modify-location
      ,x
      (lambda (,get ,set) (,set (,proc (,get)))))) )

(define-macro (ignore-values exp)
  `(##sys#call-with-values (lambda () ,exp) (lambda _ (##sys#void))) )


;;; The following is courtesy of Alex Shinn:

(define-macro (define-enum ->int ->sym . vars)
  (define (enumerate vars)
    (let loop ((n 0) (enums '()) (vars vars))
      (if (null? vars)
          (reverse enums)
          (let ((n (if (pair? (car vars))
                       (cadar vars)
                       n)))
            (loop (+ n 1)
                  (cons n enums)
                  (cdr vars))))))
  (let ((ints (enumerate vars))
        (vars (map (lambda (v) (if (pair? v) (car v) v)) vars)))
    `(begin
       ,@(map (lambda (x i) `(define-constant ,x ,i)) vars ints)
       (define (,->int sym)
         (case sym ,@(map (lambda (x i) `((,x) ,i)) vars ints) (else #f)))
       (define (,->sym int)
         (switch int ,@(map (lambda (x i) `(,i ',x)) vars ints) (else #f))))))
